// Copyright 2016 Orus.io. All rights rerund.

package cmd

import (
	"bufio"
	"fmt"
	"os"
	"os/signal"
	"runtime/debug"
	"strings"
	"syscall"

	"bitbucket.org/orus-io/xbus/fullenv_tests"
	"github.com/spf13/cobra"
)

var runCmd = &cobra.Command{
	Use:   "run",
	Short: "Run a fullenv",
	Long:  "Run a fullenv and interact on stdin/stdout",
	Run:   xbusFullenvRun,
}

func init() {
	RootCmd.AddCommand(runCmd)
	flags := runCmd.PersistentFlags()
	flags.Bool("no-prompt", false, "Disable the '>' prompt")
	v.BindPFlag("no-prompt", flags.Lookup("no-prompt"))
}

func xbusFullenvRun(cmd *cobra.Command, args []string) {
	if len(args) != 1 {
		cmd.Usage()
		return
	}

	basename, options, err := loadOptions(cmd, args[0])
	if err != nil {
		return
	}

	options.NoDBLock = true

	env := fullenv.NewFullEnv(basename, options)

	if err := env.Init(); err != nil {
		Logger.Errorf("Could not init env: %s", err)
		return
	}

	quit := make(chan struct{})

	go func() {
		signals := make(chan os.Signal, 1)
		signal.Notify(signals, os.Interrupt, syscall.SIGTERM)

		select {
		case <-signals:
			close(quit)
		case <-quit:
		}
	}()

	go inputHandler(env, quit, !v.GetBool("no-prompt"))

	<-quit

	defer env.Close()
}

func parseCmd(cmd string) (string, []string) {
	splitted := strings.Split(cmd, " ")
	return splitted[0], splitted[1:]
}

// Apply a command and returns an ouput to be sent to the user + a boolean
// indicating that the whole app should quit
func applyCmd(env *fullenv.FullEnv, cmd string, args []string) (string, bool) {
	switch cmd {
	case "client-config":
		if len(args) != 1 {
			return fmt.Sprintf("ERR: expects 1 argument (got %d)", len(args)), false
		}
		path, err := env.ClientConfig(args[0])
		if err != nil {
			return fmt.Sprintf("ERR: %s", err), false
		}
		return "OK: " + path, false
	case "wd":
		return "OK: " + env.WorkDir(), false
	case "startup":
		if err := env.Startup(); err != nil {
			Logger.Errorf("Error calling startup: %s", err)
			return fmt.Sprintf("ERR: %s", err), false
		}
		return "OK", false
	case "shutdown":
		env.Shutdown()
		return "OK", false
	case "quit":
		return "OK", true
	case "help", "?":
		fmt.Println(`Valid commands are:
    client-config <name> returns the configuration path of a client
	wd                   returns the environment work directory
    startup              start the clients
    shutdown             shutdown the clients
    quit                 shutdown everything and quit
    help, ?              this help`)
		return "OK", false
	default:
		return fmt.Sprintf("ERR: unknown command '%s'", cmd), false
	}
}

func inputHandler(env *fullenv.FullEnv, quit chan<- struct{}, prompt bool) {
	defer func() {
		if r := recover(); r != nil {
			fmt.Printf("< FATAL: %s\n", r)

			Logger.Errorf("Caught fatal error: %s\n%s", r, debug.Stack())

			close(quit)
		}
	}()

	scanner := bufio.NewScanner(os.Stdin)
	fmt.Println("< OK: " + env.WorkDir())
	if prompt {
		fmt.Print("> ")
	}
	for scanner.Scan() {
		input := scanner.Text()
		if input != "" {
			cmd, args := parseCmd(input)
			output, shouldQuit := applyCmd(env, cmd, args)
			fmt.Printf("< %s\n", output)
			if shouldQuit {
				break
			}
		}
		if prompt {
			fmt.Print("> ")
		}
	}

	if err := scanner.Err(); err != nil {
		Logger.Errorf("Error reading input: %s", err)
	}
	close(quit)
}
