// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"bufio"
	"context"
	"fmt"
	"os"
	"os/signal"
	"strconv"
	"strings"
	"sync"
	"time"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/api/control"

	"github.com/olekukonko/tablewriter"
	"github.com/spf13/cobra"
)

func contextCancelOnSignal(ctx context.Context, s ...os.Signal) (context.Context, context.CancelFunc) {
	ctx, cancel := context.WithCancel(ctx)

	c := make(chan os.Signal, 1)
	for _, sig := range s {
		signal.Notify(c, sig)
	}

	go func() {
		select {
		case <-ctx.Done():
		case <-c:
			signal.Stop(c)
			cancel()
		}
	}()

	return ctx, cancel
}

var psCmd = &cobra.Command{
	Use:   "ps",
	Short: "Process list",
	Run: func(cmd *cobra.Command, args []string) {
		filter, err := getProcessFilter(
			args, filterFlags{IncludeLiveStatuses: true})
		if err != nil {
			Logger.Errorf("Invalid filter: %s", err)
			return
		}
		c, err := GetClient()
		if err != nil {
			return
		}
		processList, err := getProcessList(c, filter, "")
		if err != nil {
			Logger.Errorf("Error: %s", err)
			return
		}
		graphList, err := c.GraphQuery("")
		if err != nil {
			Logger.Errorf("%s", err)
			return
		}
		actorList, err := c.ActorList(nil)
		if err != nil {
			Logger.Errorf("%s", err)
			return
		}
		printProcessList(processList, graphList, actorList)
	},
}

var psExportCmd = &cobra.Command{
	Use:   "export",
	Short: "Export process data",
	Run: func(cmd *cobra.Command, args []string) {
		var (
			full              = cmd.Flag("full").Value.String() == "true"
			exportEnvelope    = cmd.Flag("export-envelope").Value.String() == "true"
			exportLogs        = cmd.Flag("export-logs").Value.String() == "true"
			exportProcessLogs = cmd.Flag("export-process-logs").Value.String() == "true"

			exportEnvelopeSizeLimit int64
			exportLogLevel          api.LogLevel
		)
		if err := exportLogLevel.UnmarshalText(
			[]byte(cmd.Flag("export-logs-level").Value.String()),
		); err != nil {
			Logger.Errorf(
				"Invalid log level: '%s'. Got %s",
				cmd.Flag("export-logs-level").Value.String(), err)
			os.Exit(1)
		}

		if flag := cmd.Flag("export-envelope-size-limit"); flag.Changed {
			value, err := strconv.Atoi(flag.Value.String())
			if err != nil {
				Logger.Errorf(
					"Invalid value '%s' for --export-envelope-size-limit: %s",
					flag.Value.String(), err)
				os.Exit(1)
			}
			exportEnvelopeSizeLimit = int64(value)
		}

		c, err := GetClient()
		if err != nil {
			return
		}
		filter, err := getProcessFilter(
			args, filterFlags{IncludeAllStatuses: true})
		if err != nil {
			Logger.Errorf("Invalid filter: %s", err)
			return
		}
		list, err := getProcessList(c, filter, "")
		var pids []api.UUID
		for _, p := range list {
			pids = append(pids, p.ID)
		}
		ctx, _ := contextCancelOnSignal(
			context.Background(), os.Interrupt,
		)

		wg := sync.WaitGroup{}
		wg.Add(1)
		ch := make(chan control.OctetStreamList, 1000)

		go func() {
			for msg := range ch {
				for _, chunk := range msg.Chunks {
					fmt.Println(string(chunk.Data))
				}
			}
			wg.Done()
		}()

		if err := c.Control.Process.ExportStream(
			ctx,
			control.ProcessExportRequest{
				ProcessIDs:        pids,
				ExportEnvelope:    exportEnvelope,
				EnvelopeSizeLimit: exportEnvelopeSizeLimit,
				ExportLogs:        exportLogs,
				ExportProcessLogs: exportProcessLogs,
				ExportFull:        full,
			},
			func(ctx context.Context, msg control.OctetStreamList) {
				ch <- msg
			},
		); err != nil {
			Logger.Errorf("Error exporting stream: %s", err)
			os.Exit(1)
		}
		close(ch)
		wg.Wait()
	},
}

var psPurgeCmd = &cobra.Command{
	Use:   "purge",
	Short: "Purge process data",
	Long: `Remove all the data related to some process(es).
Defaults to all the processes terminated more than 1 month ago`,
	Run: func(cmd *cobra.Command, args []string) {
		c, err := GetClient()
		if err != nil {
			return
		}
		filter, err := getProcessFilter(
			args, filterFlags{
				IncludeTerminatedStatuses: true,
				StatusUnchangedSince:      time.Hour * 24 * 31, // Approx 1 month
			})
		if err != nil {
			Logger.Errorf("Invalid filter: %s", err)
			return
		}
		confirm := "Do you confirm the definitive deletion of %d process(es) (there is NO going back!) (yes/NO)?"
		if cmd.Flag("yes").Value.String() == "true" {
			confirm = ""
		}
		list, err := getProcessList(c, filter, confirm)
		if err != nil {
			fmt.Println(err)
			os.Exit(1)
		}
		if len(list) == 0 {
			fmt.Println("No process matched the filter")
			os.Exit(0)
		}
		var pids []api.UUID
		for _, p := range list {
			pids = append(pids, p.ID)
		}
		ctx, _ := contextCancelOnSignal(
			context.Background(), os.Interrupt,
		)
		fmt.Print("\nPurge: ")
		if err := c.ProcessPurge(ctx, pids, func(pr uint32, maxPR uint32) {
			fmt.Printf("\rPurge: %d/%d", pr, maxPR)
		}); err != nil {
			fmt.Print("\n")
			Logger.Errorf("Error: %s", err)
		} else {
			fmt.Print("\n")
			Logger.Noticef("Purged %d process(es)", len(pids))
		}
	},
}

var psPauseCmd = &cobra.Command{
	Use:   "pause",
	Short: "Pause a process",
	Run: func(cmd *cobra.Command, args []string) {
		filter, err := getProcessFilter(
			args, filterFlags{})
		if err != nil {
			Logger.Errorf("Invalid filter: %s", err)
			return
		}
		os.Exit(doCommand(
			filter, control.ProcessControlRequest_PAUSE,
			cmd.Flag("yes").Value.String() == "true",
		))
	},
}

var psResumeCmd = &cobra.Command{
	Use:   "resume",
	Short: "Resume a process",
	Run: func(cmd *cobra.Command, args []string) {
		filter, err := getProcessFilter(
			args, filterFlags{})
		if err != nil {
			Logger.Errorf("Invalid filter: %s", err)
			return
		}
		os.Exit(doCommand(
			filter, control.ProcessControlRequest_RESUME,
			cmd.Flag("yes").Value.String() == "true",
		))
	},
}

var psCancelCmd = &cobra.Command{
	Use:   "cancel",
	Short: "Cancel a process",
	Run: func(cmd *cobra.Command, args []string) {
		filter, err := getProcessFilter(
			args, filterFlags{})
		if err != nil {
			Logger.Errorf("Invalid filter: %s", err)
			return
		}
		os.Exit(doCommand(
			filter, control.ProcessControlRequest_CANCEL,
			cmd.Flag("yes").Value.String() == "true",
		))
	},
}

func doCommand(filter api.ProcessFilter, command control.ProcessControlRequest_Command, confirm bool) int {
	c, err := GetClient()
	if err != nil {
		return 1
	}
	msg := ""
	if confirm {
		msg = fmt.Sprintf(
			"Will run the command %s on %%d process(es). Do you confirm (yes/NO)?", command,
		)
	}
	list, err := getProcessList(c, filter, msg)
	if err != nil {
		fmt.Println(err)
		os.Exit(1)
	}
	if len(list) == 0 {
		fmt.Println("No process matched the filter")
		os.Exit(0)
	}
	hadErrors := false
	for _, p := range list {
		err = c.ProcessControl(p.ID, command)
		if err != nil {
			hadErrors = true
			Logger.Errorf("Process %s: %s", p.ID, err)
		}
	}
	if hadErrors {
		Logger.Errorf("Encountered errors, the command could not be applied to all processes.")
		return 1
	}
	return 0
}

func getProcessList(client *xbus.Client, filter api.ProcessFilter, confirmText string) ([]api.Process, error) {
	var processList []api.Process
	fmt.Printf("Loading process list...")
	ctx, _ := contextCancelOnSignal(
		context.Background(), os.Interrupt,
	)
	if err := client.ProcessQuery(ctx, filter, func(list []api.Process) {
		processList = append(processList, list...)
		fmt.Printf("\rLoading process list... %d", len(processList))
	}); err != nil {
		return nil, err
	}
	fmt.Printf("\rLoaded %d processes       \n", len(processList))
	if len(processList) != 0 && confirmText != "" {
		printProcessList(processList, nil, nil)
		reader := bufio.NewReader(os.Stdin)
		fmt.Printf(confirmText, len(processList))
		text, _ := reader.ReadString('\n')
		if !(text == "Y\n" || text == "y\n" || text == "yes\n") {
			return nil, nil
		}
	}
	return processList, nil
}

func printProcessList(
	processList []api.Process, graphList []api.GraphInfo, actorList []api.Actor,
) {
	var pList = processList
	if len(processList) > 1000 {
		pList = pList[:1000]
	}

	graphs := make(map[api.UUID]api.GraphInfo)
	for _, g := range graphList {
		graphs[g.ID] = g
	}
	actors := make(map[api.UUID]api.Actor)
	for _, a := range actorList {
		actors[a.ID] = a
	}
	table := tablewriter.NewWriter(os.Stdout)
	table.SetHeader([]string{"id", "graph", "emitter", "status"})

	for _, p := range pList {
		graph := p.GraphID.String()
		if g, ok := graphs[p.GraphID]; ok {
			graph = g.GetName() + " " + g.GetVersion()
		}
		actor := p.TriggerEmitterID.String()
		if a, ok := actors[p.TriggerEmitterID]; ok {
			actor = a.GetName()
		}
		table.Append([]string{
			p.ID.String(),
			graph,
			actor,
			p.Status.String(),
		})
	}
	table.Render()
	if len(processList) > 1000 {
		fmt.Printf("+ %d more processes...\n", len(processList)-1000)
	}
}

type filterFlags struct {
	StatusChangedSince        time.Duration
	StatusUnchangedSince      time.Duration
	IncludeStatusInitial      bool
	IncludeStatusRunning      bool
	IncludeStatusPaused       bool
	IncludeStatusDone         bool
	IncludeStatusError        bool
	IncludeLiveStatuses       bool
	IncludeTerminatedStatuses bool
	IncludeAllStatuses        bool
	EnvelopeIDs               string
}

var globalFlags filterFlags

func getProcessFilter(args []string, defaults filterFlags) (api.ProcessFilter, error) {
	var (
		flags  = globalFlags
		filter api.ProcessFilter
	)

	if (globalFlags == filterFlags{}) && len(args) == 0 {
		flags = defaults
	}

	if flags.StatusChangedSince != 0 {
		var t = time.Now().Add(-flags.StatusChangedSince)
		filter.StatusChangedFrom = &t
	}

	if flags.StatusUnchangedSince != 0 {
		var t = time.Now().Add(-flags.StatusUnchangedSince)
		filter.StatusChangedTo = &t
	}

	if flags.IncludeAllStatuses || flags.IncludeLiveStatuses {
		flags.IncludeStatusInitial = true
		flags.IncludeStatusRunning = true
		flags.IncludeStatusPaused = true
	}

	if flags.IncludeAllStatuses || flags.IncludeTerminatedStatuses {
		flags.IncludeStatusDone = true
		flags.IncludeStatusError = true
	}

	if flags.IncludeStatusInitial {
		filter.Status = append(filter.Status, api.Process_INITIAL)
	}

	if flags.IncludeStatusRunning {
		filter.Status = append(filter.Status, api.Process_RUNNING)
	}

	if flags.IncludeStatusPaused {
		filter.Status = append(filter.Status, api.Process_PAUSED)
	}

	if flags.IncludeStatusDone {
		filter.Status = append(filter.Status, api.Process_DONE)
	}

	if flags.IncludeStatusError {
		filter.Status = append(filter.Status, api.Process_ERROR)
	}

	if flags.EnvelopeIDs != "" {
		var sep string
		if strings.Contains(flags.EnvelopeIDs, " ") {
			sep = " "
		} else {
			sep = ","
		}
		for _, sid := range strings.Split(flags.EnvelopeIDs, sep) {
			id, err := api.UUIDFromString(sid)
			if err != nil {
				return api.ProcessFilter{}, err
			}
			filter.EnvelopeID = append(filter.EnvelopeID, id)
		}
	}

	for _, arg := range args {
		id, err := api.UUIDFromString(arg)
		if err != nil {
			return api.ProcessFilter{}, fmt.Errorf("Invalid processid: %s (%s)", arg, err)
		}
		filter.ID = append(filter.ID, id)
	}

	return filter, nil
}

func init() {
	RootCmd.AddCommand(psCmd)

	flags := psCmd.PersistentFlags()

	// A few flags to create filters
	flags.DurationVar(
		&globalFlags.StatusChangedSince,
		"status-changed-since", 0,
		"Status of the process changed at least once during the given duration before now")

	flags.DurationVar(
		&globalFlags.StatusUnchangedSince,
		"status-unchanged-since", 0,
		"Status of the process did not change during the given duration before now")

	flags.BoolVar(
		&globalFlags.IncludeStatusInitial,
		"status-initial", false,
		"Include processes in the 'initial' status")

	flags.BoolVar(
		&globalFlags.IncludeStatusRunning,
		"status-running", false,
		"Include processes in the 'running' status")

	flags.BoolVar(
		&globalFlags.IncludeStatusPaused,
		"status-paused", false,
		"Include processes in the 'paused' status")

	flags.BoolVar(
		&globalFlags.IncludeStatusDone,
		"status-done", false,
		"Include processes in the 'done' status")

	flags.BoolVar(
		&globalFlags.IncludeStatusError,
		"status-error", false,
		"Include processes in the 'error' status")

	flags.BoolVar(
		&globalFlags.IncludeLiveStatuses,
		"live", false,
		"Include processes that are not terminated")

	flags.BoolVar(
		&globalFlags.IncludeTerminatedStatuses,
		"terminated", false,
		"Include processes that are terminated")

	flags.BoolVar(
		&globalFlags.IncludeAllStatuses,
		"all", false,
		"Include processes with any status",
	)

	flags.StringVar(
		&globalFlags.EnvelopeIDs,
		"envelope-id", "",
		"Include processes that are related to one on the given envelopes",
	)

	psCmd.AddCommand(psPauseCmd)
	psCmd.AddCommand(psResumeCmd)
	psCmd.AddCommand(psCancelCmd)
	psCmd.AddCommand(psExportCmd)
	psCmd.AddCommand(psPurgeCmd)

	flags = psExportCmd.PersistentFlags()
	flags.Bool("export-envelope", false, "Export the process envelopes")
	flags.Int64("export-envelope-size-limit", 0,
		"When exporting the envelopes, limit the envelope exported size")
	flags.Bool("export-logs", false, "Export the actor logs associated to the process")
	flags.String("export-logs-level", "notice", "Log level when exporting actor logs ('notice', 'warning', 'error')")
	flags.Bool("export-process-logs", false, "Export the process logs")

	flags.Bool("full", false, "Export all the optional data (See also the --export-xxx options)")
}
