package fullenv

import (
	"crypto/x509"
	"io/ioutil"
	"path"
	"strings"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"gopkg.in/yaml.v2"

	"bitbucket.org/orus-io/go-xbus/tools"
)

func extractConfKey(filepath string, key string) interface{} {
	b, err := ioutil.ReadFile(filepath)
	if err != nil {
		panic(err)
	}
	var conf = make(map[string]interface{})
	if err := yaml.Unmarshal(b, conf); err != nil {
		panic(err)
	}

	var value interface{}
	var ok bool
	for {
		value, ok = conf[key]
		if ok {
			break
		} else if strings.Contains(key, ".") {
			splitted := strings.SplitN(key, ".", 2)
			i, ok := conf[splitted[0]]
			if ok {
				conf, ok = i.(map[string]interface{})
				if !ok {
					panic("Invalid type")
				}
				key = splitted[1]
			}
		} else {
			panic("No such key: " + key)
		}
	}
	return value
}

func extractConfKeyString(filepath string, key string) string {
	v := extractConfKey(filepath, key)
	s, ok := v.(string)
	if !ok {
		panic(key + " is not a string")
	}
	return s
}

func loadCert(t *testing.T, cert string, key string) *x509.Certificate {
	if key != "" {
		cert = extractConfKeyString(cert, key)
	}
	b, err := tools.ReadPEM("CERTIFICATE", cert)
	if err != nil {
		t.Fatal("loading cert", err)
	}
	certificate, err := x509.ParseCertificate(b)
	if err != nil {
		t.Fatal("parse cert", err)
	}
	return certificate
}

func TestCertExpirations(t *testing.T) {
	env, _ := NewFullEnvForTest(t, "certs", "certs.yaml")
	defer env.Close()

	if err := env.SetupDB(); err != nil {
		t.Fatal("SetupDB", err)
	}

	if err := env.CreateEnv(); err != nil {
		t.Fatal("CreateEnv", err)
	}

	env.Xbusd.MustRun("init")
	env.Xbusd.MustRun("key", "generate")

	env.Xbusd.MustRun("cert", "generate", "server-ca", "--expire", "1d")
	// TODO before 2051: the test will fail in 2051 because it would issue
	// an expired cert. Change the date!
	env.Xbusd.MustRun("cert", "generate", "client-ca", "--expire", "2050-12-31")

	serverCert := loadCert(
		t, path.Join(env.dirname, "certs", "xbusd", "ServerCA.crt"), "")
	serverCertExp := serverCert.NotAfter.Local()
	t.Log(serverCertExp)
	assert.Equal(t, 0, serverCertExp.Hour())
	if serverCertExp.Sub(time.Now()) > time.Hour*24 {
		t.Error("cert is not expiring in less than 24 hours")
	}

	clientCert := loadCert(
		t, path.Join(env.dirname, "certs", "xbusd", "ClientCA.crt"), "")
	clientCertExp := clientCert.NotAfter.Local()
	t.Log(clientCertExp)
	assert.Equal(t, 2050, clientCertExp.Year())
	assert.Equal(t, time.Month(12), clientCertExp.Month())
	assert.Equal(t, 31, clientCertExp.Day())
	assert.Equal(t, 0, clientCertExp.Hour())
	assert.Equal(t, 0, clientCertExp.Minute())
	assert.Equal(t, 0, clientCertExp.Second())

	if err := env.StartXbusd(); err != nil {
		t.Fatal(err)
	}

	env.XbusCtl.MustRun("key", "generate")
	env.XbusCtl.MustRun("register")
	env.Xbusd.MustRun("account", "accept", "ctl", "--cert-expire", "2m")
	env.XbusCtl.MustRun("register")

	ctlCert := loadCert(
		t, path.Join(env.dirname, "xbusctl", "xbusctl.yaml"), "tls.certificate",
	)
	ctlCertExp := ctlCert.NotAfter.Local()
	t.Log("ctlCertExp", ctlCertExp)

	if time.Now().Month() <= 10 {
		assert.Equal(t, time.Month(2), ctlCertExp.Month()-time.Now().Month())
		assert.Equal(t, time.Now().Year(), ctlCertExp.Year())
	} else {
		assert.Equal(t, time.Month(-10), ctlCertExp.Month()-time.Now().Month())
		assert.Equal(t, time.Now().Year()+1, ctlCertExp.Year())
	}
	assert.Equal(t, time.Now().Day(), ctlCertExp.Day())
	assert.Equal(t, 0, clientCertExp.Hour())
	assert.Equal(t, 0, clientCertExp.Minute())
	assert.Equal(t, 0, clientCertExp.Second())

	_, err := env.writeClientConf(env.options.Clients[0])
	if err != nil {
		t.Fatal(err)
	}
	client := NewCmdRunner(env.log, env.WorkDir()+"/client-client", "xbus-client")
	client.MustRun("register")
	env.XbusCtl.MustRun("account", "accept", "client", "--cert-expire", "10d")
	client.MustRun("register")
	clientCert = loadCert(
		t, path.Join(env.dirname, "client-client", "xbus-client.yaml"),
		"tls.certificate",
	)
	clientCertExp = clientCert.NotAfter.Local()
	assert.True(t, clientCertExp.Sub(time.Now()) < 240*time.Hour, "%s", clientCertExp.Sub(time.Now())/time.Hour/24)
}

func TestAPIKeyCLI(t *testing.T) {
	env, _ := NewFullEnvForTest(t, "apikey", "certs.yaml")
	defer env.Close()

	if err := env.Init(); err != nil {
		t.Fatal(err)
	}

	line, err := env.XbusCtl.RunReadLastLine("account", "renew-apikey", "ctl")
	if err != nil {
		t.Fatal(err)
	}

	apikey := line

	line, err = env.XbusCtl.RunReadLastLine("account", "get-apikey", "ctl")
	assert.Equal(t, apikey, line)

	line, err = env.XbusCtl.RunReadLastLine("account", "set-apikey", "ctl", "some-api-key")
	assert.Equal(t, "some-api-key", line)

	line, err = env.XbusCtl.RunReadLastLine("account", "get-apikey", "ctl")
	assert.Equal(t, "some-api-key", line)
}
