// Copyright 2016 Orus.io. All rights reserved.

package control

import (
	"context"
	"gopkg.in/yaml.v2"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/api/control"
	"bitbucket.org/orus-io/xbus/graph"
)

type graphServer struct {
	*Service
}

func (s graphServer) Query(
	ctx context.Context, req control.GraphQueryRequest,
) (resp control.GraphQueryReply, err error) {
	graphs := s.o.GraphManager.Query(req.Name, "", false)
	for _, graph := range graphs {
		resp.GraphList = append(resp.GraphList,
			api.GraphInfo{
				ID:      graph.ID,
				Name:    graph.Name,
				Version: graph.Version,
				Status:  graph.Status,
			})
	}
	return
}

func (s graphServer) Save(
	ctx context.Context, req control.GraphSaveRequest,
) (
	resp control.GraphSaveReply, err error,
) {
	var graph graph.Graph
	if err = yaml.Unmarshal([]byte(req.Graph), &graph); err != nil {
		err = api.ClientProtoError("Error decoding graph: %s", err)
		return
	}

	if req.Info.ID.IsZero() {
		graphs := s.o.GraphManager.Query(req.Info.Name, req.Info.Version, false)
		if len(graphs) > 1 {
			err = api.ClientProtoError(
				"Multiple graph matches: %s-%s", req.Info.Name, req.Info.Version)
			return
		}
		if len(graphs) == 1 {
			graph.ID = graphs[0].ID
		}
		graph.Name = req.Info.Name
		graph.Version = req.Info.Version
	} else {
		curGraph, exists := s.o.GraphManager.GetGraph(req.Info.ID)
		if exists {
			err = api.ClientProtoError(
				"No such graph: %s", req.Info.ID)
			return
		}
		graph.ID = req.Info.ID
		graph.Name = curGraph.Name
		graph.Version = curGraph.Version
	}

	for _, e := range graph.Validate() {
		resp.ValidationMessages = append(resp.ValidationMessages, e.Error())
	}

	for _, e := range graph.ValidateActors(s.o.ActorStorage) {
		resp.ValidationMessages = append(resp.ValidationMessages, e.Error())
	}

	graph, err = s.o.GraphManager.Save(graph)
	if err != nil {
		err = api.ClientProtoError("Save graph error: %s", err)
		return
	}

	var b []byte
	b, err = yaml.Marshal(map[string]interface{}{
		"nodes": graph.Nodes,
		"edges": graph.Edges,
	})
	if err != nil {
		err = api.ServerProtoError(err)
		return
	}
	resp.IsValid = len(resp.ValidationMessages) == 0
	resp.Graph = string(b)
	resp.Info = api.GraphInfo{
		ID:      graph.ID,
		Name:    graph.Name,
		Version: graph.Version,
		Status:  graph.Status,
	}
	return
}

func (s graphServer) SetStatus(ctx context.Context, req api.GraphInfo) (resp control.GraphSetStatusReply, err error) {

	var info = api.GraphInfo{ID: req.ID}
	var g graph.Graph
	if info.ID.IsZero() {
		var graphs = s.o.GraphManager.Query(req.Name, req.Version, false)
		if len(graphs) == 0 {
			err = api.ClientProtoError(
				"No such graph: %s-%s", req.Name, req.Version)
			return
		}
		if len(graphs) > 1 {
			err = api.ClientProtoError("Multiple graphs found")
			return
		}
		g = graphs[0]
		info.ID = g.ID
		info.Name = g.Name
		info.Version = g.Version
		info.Status = g.Status
	} else {
		var ok bool
		g, ok = s.o.GraphManager.GetGraph(info.ID)
		if !ok {
			err = api.ClientProtoError(
				"No such graph: %s", info.ID)
			return
		}
		info.ID = g.ID
		info.Name = g.Name
		info.Version = g.Version
		info.Status = g.Status
	}
	resp.Info = info
	if setStatusErr := s.o.GraphManager.SetStatus(g, req.Status); setStatusErr != nil {
		if list, ok := setStatusErr.(graph.ValidationErrorList); ok {
			resp.IsValid = false
			for _, e := range list {
				resp.ValidationMessages = append(resp.ValidationMessages, e.Error())
			}
		} else {
			err = api.ClientProtoError("%s", setStatusErr)
		}
	} else {
		resp.IsValid = true
		resp.Info.Status = req.Status
	}
	return
}

func (s graphServer) Get(
	ctx context.Context, req api.GraphInfo,
) (
	resp api.StringReply, err error,
) {
	var graphID = req.ID
	var result *graph.Graph
	if graphID.IsZero() {
		if req.GetName() == "" {
			err = api.ClientProtoError("Invalid graph query: must have an ID or a Name")
			return
		}
		graphs := s.o.GraphManager.Query(req.GetName(), req.GetVersion(), req.GetStatus() == api.GraphInfo_ACTIVE)
		switch len(graphs) {
		case 0:
			err = api.ClientProtoError("No such graph")
			return
		case 1:
			g := graphs[0]
			result = &g
		default:
			var active *graph.Graph
			// get the draft version if unique
			for _, g := range graphs {
				switch g.Status {
				case api.GraphInfo_DRAFT:
					if result != nil {
						err = api.ClientProtoError(
							"Graph %s has multiple drafts: please specify the version",
							req.GetName())
						return
					}
					result = &g
				case api.GraphInfo_ACTIVE:
					active = &g
				}
			}
			if result == nil && active != nil {
				result = active
			}
			if result == nil {
				err = api.ClientProtoError(
					"Graph %s has no draft of active version: please specify the version you want",
					req.GetName(),
				)
				return
			}
		}
	} else {
		g, exists := s.o.GraphManager.GetGraph(req.ID)
		if !exists {
			err = api.ClientProtoError("No such graph")
			return
		}
		result = &g
	}
	b, err := yaml.Marshal(map[string]interface{}{
		"nodes": result.Nodes,
		"edges": result.Edges,
	})
	if err != nil {
		panic(err)
	}
	resp.Value = string(b)
	return
}
