package xbus

// ActorService is a service that implements a actor behavior
type ActorService interface {
}

// StartupShutdown is for services that have a startup and a shutdown
type StartupShutdown interface {
	Startup() error
	Shutdown() error
}

// ActorProcessService is a service that has to handle process resquests
type ActorProcessService interface {
	Process(*ActorProcessingContext) error
}

// NewActorServiceFunc is a type of function that creates ActorService instances
type NewActorServiceFunc func(*Actor) ActorService

var actorServiceRegistry = make(map[string]NewActorServiceFunc)

// RegisterActorService registers a new ActorService factory function
func RegisterActorService(typename string, init NewActorServiceFunc) {
	_, ok := actorServiceRegistry[typename]
	if ok {
		panic("ActorService '" + typename + "' already registered")
	}
	actorServiceRegistry[typename] = init
}

// UnregisterActorService unregisters a ActorService
func UnregisterActorService(typename string) {
	delete(actorServiceRegistry, typename)
}

// MakeActorService returns a new ActorService, or nil if not registered
func MakeActorService(typename string, actor *Actor) ActorService {
	f, ok := actorServiceRegistry[typename]
	if !ok {
		return nil
	}
	return f(actor)
}

// RunningService holds a running service and enough context to shut it down
type RunningService struct {
	svc             ActorService
	startupShutdown StartupShutdown
	sub             *Subscription
}

// StartupService starts a service
func StartupService(actor *Actor, svc ActorService) (*RunningService, error) {
	rs := RunningService{
		svc: svc,
	}
	if ss, ok := svc.(StartupShutdown); ok {
		if err := ss.Startup(); err != nil {
			return nil, err
		}
		rs.startupShutdown = ss
	}
	if ps, ok := svc.(ActorProcessService); ok {
		sub, err := actor.Subscribe(ps.Process)
		if err != nil {
			rs.Shutdown() // TODO report errors
			return nil, err
		}
		rs.sub = sub
	}
	return &rs, nil
}

// Shutdown stops a running ActorService
func (rs *RunningService) Shutdown() error {
	if rs.sub != nil {
		rs.sub.Unsubscribe() // TODO report errors
		rs.sub = nil
	}
	if rs.startupShutdown != nil {
		// TODO report errors during shutdown
		rs.startupShutdown.Shutdown()
		rs.startupShutdown = nil
	}
	return nil
}
