// Copyright 2016 Orus.io. All rights reserved.

// Tools related to certificate management.

package tools

import (
	"crypto/ecdsa"
	"crypto/elliptic"
	"crypto/rand"
	"crypto/sha1"
	"crypto/x509"
	"crypto/x509/pkix"
	"encoding/pem"
	"errors"
	"fmt"
	"io/ioutil"
	"math/big"
	"net"
	"os"
	"strings"
	"time"
)

func fileExists(path string) bool {
	if _, err := os.Stat(path); err != nil {
		if os.IsNotExist(err) {
			return false
		}
		panic(err)
	}
	return true
}

// CreatePrivateKey generates a private key and returns it, PEM-encoded.
func CreatePrivateKey() (privKeyData []byte, err error) {
	curve := elliptic.P521()
	key, err := ecdsa.GenerateKey(curve, rand.Reader)
	if err != nil {
		return nil, err
	}

	return PKeyToPEM(key)
}

// PKeyToPEM serialize a key to PEM encoding
func PKeyToPEM(key *ecdsa.PrivateKey) ([]byte, error) {
	derBytes, err := x509.MarshalECPrivateKey(key)
	if err != nil {
		return nil, err
	}
	return pem.EncodeToMemory(&pem.Block{Type: "EC PRIVATE KEY", Bytes: derBytes}), nil
}

// ReadPEM loads bytes from a PEM content or a file containing a PEM encoded block
func ReadPEM(blockType string, pemOrPath string) ([]byte, error) {
	var (
		bytes []byte
		err   error
	)
	if strings.Contains(pemOrPath, "-----BEGIN") {
		if !strings.Contains(pemOrPath, "-----BEGIN "+blockType+"-----") {
			return nil, fmt.Errorf(
				"Invalid pem block type. Expected %s, got:\n%s",
				blockType, pemOrPath)
		}
		bytes = []byte(pemOrPath)
	} else if fileExists(pemOrPath) {
		bytes, err = ioutil.ReadFile(pemOrPath)
		if err != nil {
			return nil, err
		}
	} else {
		return nil, fmt.Errorf("Not a file nor a PEM block: %s", pemOrPath)
	}
	block, _ := pem.Decode(bytes)
	if block == nil {
		return nil, errors.New("Could not decode PEM data")
	}
	return block.Bytes, nil
}

// ReadPEMCertificate loads a x509.Certificate from a PEM file or content
func ReadPEMCertificate(pemOrPath string) (*x509.Certificate, error) {
	bytes, err := ReadPEM("CERTIFICATE", pemOrPath)
	if err != nil {
		return nil, err
	}
	return x509.ParseCertificate(bytes)
}

// ReadPEMPrivateKey loads a x509.Certificate from a PEM file or content
func ReadPEMPrivateKey(pemOrPath string) (*ecdsa.PrivateKey, error) {
	bytes, err := ReadPEM("EC PRIVATE KEY", pemOrPath)
	if err != nil {
		return nil, err
	}
	return x509.ParseECPrivateKey(bytes)
}

// CreateServerRootCA creates a self-signed X.509
// certificate used by the Xbus server.
func CreateServerRootCA(
	priv interface{}, notbefore, notafter time.Time, hostDefs string,
) (certData []byte, err error) {
	// TODO Allow configuring the following settings.
	return CreateCert(
		priv, true, x509.ExtKeyUsageServerAuth, nil, nil,
		notbefore, notafter,
		hostDefs, []string{"Xbus"}, "XbusRootCA",
	)
}

// CreateClientCA creates a self-signed X.509 CA for client connections.
func CreateClientCA(
	priv interface{}, notbefore, notafter time.Time,
) (certData []byte, err error) {
	// TODO Allow configuring the following settings.
	return CreateCert(
		priv, true, x509.ExtKeyUsageClientAuth, nil, nil,
		notbefore, notafter,
		"localhost", []string{"Xbus"}, "XbusClientCA",
	)
}

// CreateCSR is used when creating CSRs.
// you need to provide a private Key
func CreateCSR(
	priv interface{}, hostDefs string, certSubject pkix.Name,
) (derBytes []byte, err error) {

	ipList, dnsList := parseCertHosts(hostDefs)

	template := x509.CertificateRequest{
		Subject:     certSubject,
		IPAddresses: ipList,
		DNSNames:    dnsList,
	}

	return x509.CreateCertificateRequest(rand.Reader, &template, priv)
}

// CreateCert creates either an X.509 certificate (self-signed or
// to be signed via the specified CA private key) or a PKCS #10 CSR.
func CreateCert(
	priv interface{}, ca bool,
	extKeyUsage x509.ExtKeyUsage,
	signerCertData []byte, signerPrivKey []byte,
	notbefore, notafter time.Time, hostDefs string,
	organization []string, commonName string,
) (certData []byte, err error) {

	// Heavily based on go's crypto/tls/generate_cert.go.

	certSubject := pkix.Name{
		Organization: organization,
		CommonName:   commonName,
	}

	serialNumberLimit := new(big.Int).Lsh(big.NewInt(1), 128)
	serialNumber, err := rand.Int(rand.Reader, serialNumberLimit)
	if err != nil {
		return nil, fmt.Errorf("Failed to generate serial number: %s", err)
	}

	ipList, dnsList := parseCertHosts(hostDefs)

	template := x509.Certificate{
		SerialNumber: serialNumber,
		Subject:      certSubject,
		IPAddresses:  ipList,
		DNSNames:     dnsList,
		NotBefore:    notbefore,
		NotAfter:     notafter,

		KeyUsage:    x509.KeyUsageKeyEncipherment | x509.KeyUsageDigitalSignature,
		ExtKeyUsage: []x509.ExtKeyUsage{extKeyUsage},

		BasicConstraintsValid: true,
	}

	if ca {
		template.KeyUsage |= x509.KeyUsageCertSign
		template.IsCA = true
	}

	var signerCert x509.Certificate

	if signerCertData == nil {
		// No signer specified; go for a self-signed certificate.
		signerCert = template

	} else {
		// Build a certificate structure for the signer.
		signerCertBlock, _ := pem.Decode(signerCertData)
		signerCertBuilt, err := x509.ParseCertificate(signerCertBlock.Bytes)
		if err != nil {
			return nil, fmt.Errorf(
				"Failed to load the specified signer certificate: %s",
				err,
			)
		}
		signerCert = *signerCertBuilt
	}

	// When a private key is provided, use it;
	// otherwise, use the generated one.
	var signerPriv interface{}
	if signerPrivKey != nil {

		signerPrivBlock, _ := pem.Decode(signerPrivKey)
		signerPriv, err = x509.ParseECPrivateKey(signerPrivBlock.Bytes)
		if err != nil {
			return nil, fmt.Errorf(
				"Failed to load the specified signer private key: %s",
				err,
			)
		}

	} else {
		signerPriv = priv
	}

	certBytes, err := x509.CreateCertificate(
		rand.Reader, &template,
		&signerCert, &priv.(*ecdsa.PrivateKey).PublicKey, signerPriv,
	)

	if err != nil {
		return nil, err
	}

	return pem.EncodeToMemory(&pem.Block{Type: "CERTIFICATE", Bytes: certBytes}), nil
}

// parseCertHosts is an internal function that splits a list of
// hosts between IPs / DNSs, ready to be put into a one
// of the "x509" template structs.
func parseCertHosts(hostDefs string) (ipList []net.IP, dnsList []string) {
	hosts := strings.Split(hostDefs, ",")
	for _, host := range hosts {
		if ip := net.ParseIP(host); ip != nil {
			ipList = append(ipList, ip)
		} else {
			dnsList = append(dnsList, host)
		}
	}
	return ipList, dnsList
}

// CreateCertificateFromCSR creates a certificate using a csr as a template.
func CreateCertificateFromCSR(
	csr []byte, parent *x509.Certificate, priv interface{}, notAfter time.Time,
) ([]byte, error) {
	cr, err := x509.ParseCertificateRequest(csr)
	if err != nil {
		return nil, err
	}
	err = cr.CheckSignature()
	if err != nil {
		return nil, err
	}
	serialNumberLimit := new(big.Int).Lsh(big.NewInt(1), 128)
	serialNumber, err := rand.Int(rand.Reader, serialNumberLimit)
	if err != nil {
		return nil, err
	}
	template := x509.Certificate{
		NotBefore:    time.Now(),
		NotAfter:     notAfter,
		SerialNumber: serialNumber,
		Subject:      cr.Subject,
		IPAddresses:  cr.IPAddresses,
		DNSNames:     cr.DNSNames,
		ExtKeyUsage:  []x509.ExtKeyUsage{x509.ExtKeyUsageClientAuth},
	}

	return x509.CreateCertificate(rand.Reader, &template, parent, cr.PublicKey, priv)
}

// GetCSRHash computes a SHA-1 hash of the specified certificate signing request; the output is
// prettyfied (2-char split). The hash is based on the DER data of the CSR.
// @param decodePEM: Whether the data is given as PEM-encoded data, in which case the DER data will
// be extracted from it.
func GetCSRHash(csrDataStr string, decodePEM bool) string {
	if len(csrDataStr) == 0 {
		return ""
	}

	// Convert to a byte array.
	csrData := []byte(csrDataStr)

	if decodePEM {
		// Extract the DER data from the PEM-encoded CSR.
		csrBlock, _ := pem.Decode(csrData)
		if csrBlock.Type != "CERTIFICATE REQUEST" {
			return ""
		}
		csrData = csrBlock.Bytes
	}

	// Compute a SHA-1 hash of the CSR.
	hash := fmt.Sprintf("%X", sha1.Sum(csrData))

	// Split into blocks of 2 chars.
	var blocks []string
	for i := 0; i < len(hash); i += 2 {
		blocks = append(blocks, fmt.Sprintf("%s", hash[i:i+2]))
	}
	return strings.Join(blocks, " ")
}
