package tools

import (
	"time"

	"github.com/nats-io/go-nats"
	"github.com/nats-io/nuid"
	"github.com/nats-rpc/gogo-nrpc"
)

// NewClientConn returns a CustomInboxPrefixConn for a client
func NewClientConn(c *nats.Conn, ClientID string) *CustomInboxPrefixConn {
	return &CustomInboxPrefixConn{
		Conn:        c,
		InboxPrefix: nats.InboxPrefix + ClientID + ".in.",
	}
}

// NewServerToClientConn returns a CustomInboxPrefixConn for a server talking to a given client
func NewServerToClientConn(c *nats.Conn, ClientID string) *CustomInboxPrefixConn {
	return &CustomInboxPrefixConn{
		Conn:        c,
		InboxPrefix: nats.InboxPrefix + ClientID + ".out.",
	}
}

// CustomInboxPrefixConn is a nats.Conn that uses a custom inbox prefix
type CustomInboxPrefixConn struct {
	*nats.Conn

	InboxPrefix string
}

// Request overrides nats.Conn.Request and uses a custom reply subject
func (c CustomInboxPrefixConn) Request(subject string, data []byte, timeout time.Duration) (*nats.Msg, error) {
	return RequestWithInbox(
		c.Conn, subject, c.InboxPrefix+nuid.Next(),
		data, timeout)
}

// RequestWithInbox is a nats.Conn.Request that takes the reply instead of
// calculating it automatically
func RequestWithInbox(nc *nats.Conn, subject, inbox string, data []byte, timeout time.Duration) (*nats.Msg, error) {
	ch := make(chan *nats.Msg, nats.RequestChanLen)

	s, err := nc.ChanSubscribe(inbox, ch)
	if err != nil {
		return nil, err
	}
	s.AutoUnsubscribe(1)
	defer s.Unsubscribe()

	err = nc.PublishRequest(subject, inbox, data)
	if err != nil {
		return nil, err
	}
	return s.NextMsg(timeout)
}

// NewWhoAmIConn creates a new WhoAmIConn
func NewWhoAmIConn(c *nats.Conn) WhoAmIConn {
	return WhoAmIConn{c}
}

// WhoAmIConn is a conn with a very specific reply
type WhoAmIConn struct {
	*nats.Conn
}

// Request performs a custom REQ/REP for the 'whoami' call that have
// a specific inbox subject pattern
func (c WhoAmIConn) Request(subject string, data []byte, timeout time.Duration) (*nats.Msg, error) {
	ch := make(chan *nats.Msg, nats.RequestChanLen)

	inbox := subject + ".reply"
	s, err := c.Conn.ChanSubscribe(inbox, ch)
	if err != nil {
		return nil, err
	}
	s.AutoUnsubscribe(1)
	defer s.Unsubscribe()

	err = c.Conn.PublishRequest(subject, inbox, data)
	if err != nil {
		return nil, err
	}
	return s.NextMsg(timeout)
}

func init() {
	nrpc.GetReplyInbox = func(nc nrpc.NatsConn) string {
		if conn, ok := nc.(*CustomInboxPrefixConn); ok {
			return conn.InboxPrefix + nuid.Next()
		}
		return nats.NewInbox()
	}
}
